from pynamodax.attributes import (
    Attribute,
    BinaryAttribute,
    BinarySetAttribute,
    DiscriminatorAttribute,
    JSONAttribute,
    TTLAttribute,
    UTCDateTimeAttribute,
)

SERIALIZABLE_TYPES = (BinaryAttribute, BinarySetAttribute, DiscriminatorAttribute, JSONAttribute)
ENCODER_MAPPING = {
    SERIALIZABLE_TYPES: lambda attr, data: attr.serialize(data),
    TTLAttribute: lambda _, data: data.timestamp(),
    UTCDateTimeAttribute: lambda _, data: data.isoformat(),
}


class PrimitiveAttributeEncoder:
    @staticmethod
    def encode(attr: Attribute, data):
        for types, callable in ENCODER_MAPPING.items():
            if isinstance(attr, types):
                return callable(attr, data)
        return data



"""
Utils
"""
import json
from typing import Any
from typing import Dict

from pynamodax.constants import BINARY
from pynamodax.constants import BINARY_SET
from pynamodax.constants import BOOLEAN
from pynamodax.constants import LIST
from pynamodax.constants import MAP
from pynamodax.constants import NULL
from pynamodax.constants import NUMBER
from pynamodax.constants import NUMBER_SET
from pynamodax.constants import STRING
from pynamodax.constants import STRING_SET


def attribute_value_to_json(attribute_value: Dict[str, Any]) -> Any:
    attr_type, attr_value = next(iter(attribute_value.items()))
    if attr_type == LIST:
        return [attribute_value_to_json(v) for v in attr_value]
    if attr_type == MAP:
        return {k: attribute_value_to_json(v) for k, v in attr_value.items()}
    if attr_type == NULL:
        return None
    if attr_type in {BINARY, BINARY_SET, BOOLEAN, STRING, STRING_SET}:
        return attr_value
    if attr_type == NUMBER:
        return json.loads(attr_value)
    if attr_type == NUMBER_SET:
        return [json.loads(v) for v in attr_value]
    raise ValueError("Unknown attribute type: {}".format(attr_type))


def json_to_attribute_value(value: Any) -> Dict[str, Any]:
    if value is None:
        return {NULL: True}
    if value is True or value is False:
        return {BOOLEAN: value}
    if isinstance(value, (int, float)):
        return {NUMBER: json.dumps(value)}
    if isinstance(value, str):
        return {STRING: value}
    if isinstance(value, list):
        return {LIST: [json_to_attribute_value(v) for v in value]}
    if isinstance(value, dict):
        return {MAP: {k: json_to_attribute_value(v) for k, v in value.items()}}
    raise ValueError("Unknown value type: {}".format(type(value).__name__))



"""
Type-annotates the private botocore APIs that we're currently relying on.
"""
from typing import Any, Dict

import botocore.client
import botocore.credentials
import botocore.endpoint
import botocore.hooks
import botocore.model
import botocore.signers


class BotocoreEndpointPrivate(botocore.endpoint.Endpoint):
    _event_emitter: botocore.hooks.HierarchicalEmitter


class BotocoreRequestSignerPrivate(botocore.signers.RequestSigner):
    _credentials: botocore.credentials.Credentials


class BotocoreBaseClientPrivate(botocore.client.BaseClient):
    _endpoint: BotocoreEndpointPrivate
    _request_signer: BotocoreRequestSignerPrivate
    _service_model: botocore.model.ServiceModel

    def _convert_to_request_dict(self, api_params: Dict[str, Any], operation_model: botocore.model.OperationModel, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        ...


from typing import Any, Dict, List, Type, TypeVar, Union

from pynamodax.attributes import (
    Attribute,
    AttributeContainer,
    DiscriminatorAttribute,
    DynamicMapAttribute,
    ListAttribute,
    MapAttribute,
)

from pynamodax.encoding.primitive_attribute_decoder import PrimitiveAttributeDecoder

AC = TypeVar("AC", bound=AttributeContainer)


class Decoder:
    def decode(self, type: Type, data: Dict[str, Any]):
        return self.decode_container(type, data)

    def decode_container(self, type: Type[AC], data: Dict[str, Any]) -> AC:
        attributes = {}
        cls = self.polymorphisize(type, data)
        for name, attr in cls.get_attributes().items():
            if name in data:
                attributes[name] = self.decode_attribute(attr, data[name])
        return cls(**attributes)

    def decode_attribute(self, attr: Attribute, data):
        if isinstance(attr, ListAttribute):
            return self.decode_list(attr, data)
        elif type(attr) == MapAttribute or issubclass(type(attr), MapAttribute):
            return self.decode_map(attr, data)
        else:
            return PrimitiveAttributeDecoder.decode(attr, data)

    def decode_list(self, attr: ListAttribute, data: List) -> List:
        return [self.decode_attribute(self.coerce(attr.element_type), value) for value in data]

    def coerce(self, element_type: Union[Type[Attribute], None]) -> Attribute:
        return (element_type or Attribute)()

    def decode_map(self, attr: MapAttribute, data: Dict[str, Any]):
        if isinstance(attr, MapAttribute) or issubclass(type(attr), MapAttribute):
            return data
        elif isinstance(attr, DynamicMapAttribute) or issubclass(type(attr), DynamicMapAttribute):
            return self.decode_dynamic_map(attr, data)
        else:
            return self.decode_container(type(attr), data)

    def decode_dynamic_map(self, attr: DynamicMapAttribute, data: Dict[str, Any]):
        decoded = {}
        cls = self.polymorphisize(type(attr), data)
        attributes = attr.get_attributes()
        for name, value in data.items():
            if name in attributes:
                decoded[name] = self.decode_attribute(attributes[name], value)
            else:
                decoded[name] = value
        return cls(**decoded)

    def polymorphisize(self, instance_type: Type[AC], data: Dict[str, Any]) -> Type[AC]:
        for name, attr in instance_type.get_attributes().items():
            if isinstance(attr, DiscriminatorAttribute):
                return attr.deserialize(data.pop(name))
        return instance_type



from typing import Any, Dict, List

from pynamodax.attributes import (
    Attribute,
    AttributeContainer,
    DynamicMapAttribute,
    ListAttribute,
    MapAttribute,
)

from pynamodax.encoding.primitive_attribute_encoder import PrimitiveAttributeEncoder


class Encoder:
    def encode(self, instance) -> Dict[str, Any]:
        return self.encode_container(instance)

    def encode_container(self, container: AttributeContainer) -> Dict[str, Any]:
        encoded = {}
        for name, attr in container.get_attributes().items():
            value = getattr(container, name)
            if value is not None:
                encoded[name] = self.encode_attribute(attr, value)
        return encoded

    def encode_attribute(self, attr: Attribute, data: Any):
        if isinstance(attr, ListAttribute):
            return self.encode_list(attr, data)
        elif type(attr) == MapAttribute or issubclass(type(attr), MapAttribute):
            return self.encode_map(attr, data)
        else:
            return PrimitiveAttributeEncoder.encode(attr, data)

    def encode_list(self, attr: ListAttribute, data: List) -> List:
        element_attr = (attr.element_type or Attribute)()
        return [self.encode_attribute(element_attr, value) for value in data]

    def encode_map(self, attr: MapAttribute, data: MapAttribute) -> Dict:
        if isinstance(attr, MapAttribute) or issubclass(type(attr), MapAttribute):
            return {name: data[name] for name in data}
        elif isinstance(attr, DynamicMapAttribute) or issubclass(type(attr), DynamicMapAttribute):
            return self.encode_dynamic_map(attr, data)
        else:
            return self.encode_container(data)

    def encode_dynamic_map(self, attr: DynamicMapAttribute, data: MapAttribute) -> Dict:
        encoded = {}
        attributes = attr.get_attributes()
        for name in data:
            value = getattr(data, name)
            if name in attributes:
                encoded[name] = self.encode_attribute(attributes[name], value)
            else:
                encoded[name] = value
        return encoded



from typing import Dict, List

OP_WRITE = {
    'PutItem': 'put_item',
    'DeleteItem': 'delete_item',
    'UpdateItem': 'update_item',
    'BatchWriteItem': 'batch_write_item',
}

OP_READ = {
    'GetItem': 'get_item',
    'Scan': 'scan',
    'BatchGetItem': 'batch_get_item',
    'Query': 'query',
}

OP_NAME_TO_METHOD = OP_WRITE.copy()
OP_NAME_TO_METHOD.update(OP_READ)


class DaxClient(object):

    def __init__(self, endpoints: List[str], region_name: str):
        from amazondax import AmazonDaxClient

        self.connection = AmazonDaxClient(
            endpoints=endpoints,
            region_name=region_name
        )

    def dispatch(self, operation_name: str, operation_kwargs: Dict):
        method = getattr(self.connection, OP_NAME_TO_METHOD[operation_name])
        return method(**operation_kwargs)



from setuptools import setup, find_packages


install_requires = [
    'botocore>=1.12.54',
    'typing-extensions>=3.7; python_version<"3.8"'
]

setup(
    name='pynamodax',
    version=__import__('pynamodax').__version__,
    packages=find_packages(exclude=('examples', 'tests', 'tests.integration',)),
    url='http://jlafon.io/pynamodb.html',
    project_urls={
        'Source': 'https://github.com/pynamodb/PynamoDB',
    },
    author='Jharrod LaFon',
    author_email='jlafon@eyesopen.com',
    description='A Pythonic Interface to DynamoDB',
    long_description=open('README.rst').read(),
    zip_safe=False,
    license='MIT',
    keywords='python dynamodb amazon dax',
    python_requires=">=3.7",
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
    ],
    extras_require={
        'signals': ['blinker>=1.3,<2.0'],
        'dax': ['amazon-dax-client>=2.0.0,<3.0.0']
    },
    package_data={'pynamodax': ['py.typed']},
)



from datetime import datetime, timezone

from pynamodax.attributes import (
    Attribute,
    BinaryAttribute,
    BinarySetAttribute,
    JSONAttribute,
    NumberSetAttribute,
    TTLAttribute,
    UnicodeSetAttribute,
    UTCDateTimeAttribute,
)

DECODER_MAPPING = {
    (BinaryAttribute, BinarySetAttribute, JSONAttribute): lambda attr, data: attr.deserialize(data),
    (NumberSetAttribute, UnicodeSetAttribute): lambda _, data: set(data),
    TTLAttribute: lambda _, data: datetime.fromtimestamp(data, tz=timezone.utc),
    UTCDateTimeAttribute: lambda _, data: datetime.fromisoformat(data),
}


class PrimitiveAttributeDecoder:
    @staticmethod
    def decode(attr: Attribute, data):
        for types, callable in DECODER_MAPPING.items():
            if isinstance(attr, types):
                return callable(attr, data)
        return data
